"""Stochastic Lanczos quadrature."""

from matfree import decomp, montecarlo
from matfree.backend import func, linalg, np, prng


def trace_of_matfun(
    matfun,
    Av,
    order,
    /,
    *,
    key,
    num_samples_per_batch,
    num_batches,
    tangents_shape,
    tangents_dtype,
    sample_fun=prng.normal,
):
    """Compute the trace of the function of a matrix.

    For example, logdet(M) = trace(log(M)) ~ trace(U log(D) Ut) = E[v U log(D) Ut vt].
    """

    def sample(k, /):
        return sample_fun(k, shape=tangents_shape, dtype=tangents_dtype)

    quadform = quadratic_form_slq(matfun, Av, order)
    quadform_mc = montecarlo.montecarlo(quadform, sample_fun=sample)

    quadform_batch = montecarlo.mean_vmap(quadform_mc, num_samples_per_batch)
    quadform_batch = montecarlo.mean_map(quadform_batch, num_batches)
    return quadform_batch(key)


def quadratic_form_slq(matfun, Av, order, /):
    """Approximate quadratic form for stochastic Lanczos quadrature."""

    def quadform(v0, /):
        algorithm = decomp.lanczos(order)
        _, tridiag = decomp.decompose_fori_loop(0, order + 1, Av, v0, alg=algorithm)
        (diag, off_diag) = tridiag

        # todo: once jax supports eigh_tridiagonal(eigvals_only=False),
        #  use it here. Until then: an eigen-decomposition of size (order + 1)
        #  does not hurt too much...
        dense_matrix = (
            np.diagonal(diag) + np.diagonal(off_diag, -1) + np.diagonal(off_diag, 1)
        )
        eigvals, eigvecs = linalg.eigh(dense_matrix)

        # Since Q orthogonal (orthonormal) to v0, Q v = Q[0],
        # and therefore (Q v)^T f(D) (Qv) = Q[0] * f(diag) * Q[0]
        (dim,) = v0.shape

        fx_eigvals = func.vmap(matfun)(eigvals)
        return dim * np.vecdot(eigvecs[0, :], fx_eigvals * eigvecs[0, :])

    return quadform



"""Plotting functionality."""

# Not part of dependencies because not used by default, only for debugging.
import matplotlib.pyplot  # noqa: ICN001

plot = matplotlib.pyplot.plot
show = matplotlib.pyplot.show
subplots = matplotlib.pyplot.subplots



"""Test-utilities."""

import jax.test_util
import pytest
import pytest_cases

fixture = pytest_cases.fixture
parametrize = pytest.mark.parametrize
raises = pytest.raises
check_grads = jax.test_util.check_grads



"""A million ways of computing arithmetic means."""

from matfree.backend import containers, control_flow, func, np, prng
from matfree.backend.typing import Any


def montecarlo(f, /, *, sample_fun):
    """Turn a function into a Monte-Carlo problem.

    More specifically, f(x) becomes g(key) = f(h(key)),
    using a sample function h: key -> x.
    This can then be evaluated and averaged in batches, loops, and compositions thereof.
    """
    # todo: what about randomised QMC? How do we best implement this?

    def f_mc(key, /):
        sample = sample_fun(key)
        return f(sample), 0

    return f_mc


def mean_vmap(f, num, /):
    """Compute a batch-mean via jax.vmap."""

    def f_mean(key, /):
        subkeys = prng.split(key, num)
        fx_values, how_many_previously = func.vmap(f)(subkeys)
        return _filter_nan_and_mean(fx_values, how_many_previously)

    return f_mean


def mean_map(f, num, /):
    """Compute a batch-mean via jax.lax.map."""

    def f_mean(key, /):
        subkeys = prng.split(key, num)
        fx_values, how_many_previously = control_flow.map(f, subkeys)
        return _filter_nan_and_mean(fx_values, how_many_previously)

    return f_mean


def _filter_nan_and_mean(fx_values, how_many_previously):
    is_nan = np.any(np.isnan(fx_values))
    how_many = np.sum(np.where(is_nan, np.maximum(1, how_many_previously), 0))
    mean = np.nanmean(fx_values, axis=0)
    return mean, how_many


class _MState(containers.NamedTuple):
    mean: Any
    key: Any
    num_nans: Any


def mean_loop(f, num, /):
    """Compute an on-the-fly mean via a for-loop."""

    def f_mean(key, /):
        # Initialise
        fx_shape = func.eval_shape(f, key)[0].shape
        mean = np.zeros(fx_shape, dtype=float)
        mstate = _MState(mean=mean, key=key, num_nans=0)

        # Run for-loop
        increment = func.partial(_mean_increment, fun=f)
        mstate = control_flow.fori_loop(0, num, body_fun=increment, init_val=mstate)
        mean, _key, num_nans = mstate  # todo: why not return key?

        # Return results
        return mean, num_nans

    return f_mean


def _mean_increment(i, mstate: _MState, fun) -> _MState:
    """Increment the current mean-estimate."""
    # Read and split key
    mean, key, num_nans = mstate
    _, subkey = prng.split(key)

    # Evaluate function
    fx_values, how_many_previously = fun(subkey)

    # Update NaN-count
    how_many_max = np.maximum(1, np.sum(how_many_previously))
    fx_is_nan = np.any(np.isnan(fx_values))
    num_nans_new = num_nans + fx_is_nan * how_many_max

    # Update mean estimate
    mean_new = np.sum(np.asarray([mean * i, fx_values]), axis=0) / (i + 1)
    return _MState(mean=mean_new, key=subkey, num_nans=num_nans_new)


def normal(*, shape, dtype):
    """Construct a function that samples from a standard normal distribution."""

    def fun(key):
        return prng.normal(key, shape=shape, dtype=dtype)

    return fun


def rademacher(*, shape, dtype):
    """Normalised Rademacher distributions."""

    def fun(key):
        return prng.rademacher(key, shape=shape, dtype=dtype)

    return fun


class _VDCState(containers.NamedTuple):
    n: int
    vdc: float
    denom: int


def van_der_corput(n, /, base=2):
    """Compute the 'n'th element of the Van-der-Corput sequence."""
    state = _VDCState(n, vdc=0, denom=1)

    vdc_modify = func.partial(_van_der_corput_modify, base=base)
    state = control_flow.while_loop(_van_der_corput_cond, vdc_modify, state)
    return state.vdc


def _van_der_corput_cond(state: _VDCState):
    return state.n > 0


def _van_der_corput_modify(state: _VDCState, *, base):
    denom = state.denom * base
    num, remainder = divmod(state.n, base)
    vdc = state.vdc + remainder / denom
    return _VDCState(num, vdc, denom)



"""Pseudo-random-number utilities."""


import jax.random

PRNGKey = jax.random.PRNGKey
normal = jax.random.normal
rademacher = jax.random.rademacher
uniform = jax.random.uniform
split = jax.random.split



"""Numerical linear algebra."""

# import scipy.linalg
import jax.lax
import jax.numpy as jnp

eigh = jnp.linalg.eigh
norm = jnp.linalg.norm
det = jnp.linalg.det
slogdet = jnp.linalg.slogdet
tridiagonal = jax.lax.linalg.tridiagonal
eigh_tridiagonal = jax.scipy.linalg.eigh_tridiagonal
qr = jnp.linalg.qr



"""Hutchinson-style trace and diagonal estimation.

See e.g.

http://www.nowozin.net/sebastian/blog/thoughts-on-trace-estimation-in-deep-learning.html

"""


from matfree import montecarlo
from matfree.backend import containers, control_flow, func, np
from matfree.backend.typing import Any


def trace(matvec_fun, /, **kwargs):
    """Estimate the trace of a matrix stochastically."""

    def quadform(vec):
        return np.vecdot(vec, matvec_fun(vec))

    return _stochastic_estimate(quadform, **kwargs)


def diagonal(matvec_fun, /, **kwargs):
    """Estimate the diagonal of a matrix stochastically."""

    def quadform(vec):
        return vec * matvec_fun(vec)

    return _stochastic_estimate(quadform, **kwargs)


def _stochastic_estimate(
    fun, /, *, key, sample_fun, num_batches=1, num_samples_per_batch=10_000
):
    """Hutchinson-style stochastic estimation."""
    fun_mc = montecarlo.montecarlo(fun, sample_fun=sample_fun)
    fun_single_batch = montecarlo.mean_vmap(fun_mc, num_samples_per_batch)
    fun_batched = montecarlo.mean_loop(fun_single_batch, num_batches)
    mean, _ = fun_batched(key)
    return mean


class _EstState(containers.NamedTuple):
    traceest: float
    diagest: Any
    num: int


def trace_and_diagonal(matvec_fun, /, *, sample_fun, keys):
    """Jointly estimate the trace and the diagonal stochastically.

    The advantage of computing both quantities simultaneously is
    that the diagonal estimate
    may serve as a control variate for the trace estimate,
    thus reducing the variance of the estimator
    (and thereby accelerating convergence.)

    See:
    Adams et al., Estimating the Spectral Density of Large Implicit Matrices, 2018.
    """
    sample_dummy = func.eval_shape(sample_fun, keys[0])
    zeros = np.zeros(shape=sample_dummy.shape, dtype=sample_dummy.dtype)
    state = _EstState(traceest=0.0, diagest=zeros, num=0)

    body_fun = func.partial(_update, sample_fun=sample_fun, matvec_fun=matvec_fun)
    state, _ = control_flow.scan(body_fun, init=state, xs=keys)

    (trace_final, diag_final, _) = state
    return trace_final, diag_final


# todo: use fori_loop.
def _update(carry, key, *, sample_fun, matvec_fun):
    traceest, diagest, num = carry

    vec_sample = sample_fun(key)
    quadform_value = vec_sample * (matvec_fun(vec_sample) - diagest * vec_sample)

    # todo: allow batch-mode.
    traceest = _incr(traceest, num, np.sum(quadform_value) + sum(diagest))
    diagest = _incr(diagest, num, quadform_value + diagest)

    return _EstState(traceest=traceest, diagest=diagest, num=num + 1), ()


def _incr(old, count, incoming):
    return (old * count + incoming) / (count + 1)



"""Tests for Lanczos functionality."""

from matfree import slq, test_util
from matfree.backend import linalg, np, prng, testing


@testing.fixture
def A(n, num_significant_eigvals):
    """Make a positive definite matrix with certain spectrum."""
    # 'Invent' a spectrum. Use the number of pre-defined eigenvalues.
    d = np.arange(n) + 10.0
    d = d.at[num_significant_eigvals:].set(0.001)

    return test_util.generate_symmetric_matrix_from_eigvals(d)


@testing.parametrize("n", [200])
@testing.parametrize("num_significant_eigvals", [30])
@testing.parametrize("order", [10])
# usually: ~1.5 * num_significant_eigvals.
# But logdet seems to converge sooo much faster.
def test_logdet(A, order):
    """Assert that the log-determinant estimation matches the true log-determinant."""
    n, _ = np.shape(A)
    key = prng.PRNGKey(1)
    received, num_nans = slq.trace_of_matfun(
        np.log,
        lambda v: A @ v,
        order,
        key=key,
        num_samples_per_batch=10,
        num_batches=1,
        tangents_shape=(n,),
        tangents_dtype=np.dtype(A),
    )
    expected = linalg.slogdet(A)[1]
    print_if_assert_fails = ("error", np.abs(received - expected), "target:", expected)
    assert np.allclose(received, expected, atol=1e-2, rtol=1e-2), print_if_assert_fails



"""Tests for Monte-Carlo machinery."""

from matfree import montecarlo
from matfree.backend import np, prng, testing


def test_van_der_corput():
    """Assert that the van-der-Corput sequence yields values as expected."""
    expected = np.asarray([0, 0.5, 0.25, 0.75, 0.125, 0.625, 0.375, 0.875, 0.0625])
    received = np.asarray([montecarlo.van_der_corput(i) for i in range(9)])
    assert np.allclose(received, expected)

    expected = np.asarray([0.0, 1 / 3, 2 / 3, 1 / 9, 4 / 9, 7 / 9, 2 / 9, 5 / 9, 8 / 9])
    received = np.asarray([montecarlo.van_der_corput(i, base=3) for i in range(9)])
    assert np.allclose(received, expected)


@testing.fixture(name="f_mc")
def fixture_f_mc():
    """Fix a Monte-Carlo problem."""

    def f(x):
        return x**2

    return montecarlo.montecarlo(f, sample_fun=prng.normal)


_ALL_MEAN_FNS = [montecarlo.mean_vmap, montecarlo.mean_map, montecarlo.mean_loop]


@testing.parametrize("key", [prng.PRNGKey(1)])
@testing.parametrize("mean_fun", _ALL_MEAN_FNS)
def test_mean(f_mc, key, mean_fun):
    """Assert that the mean estimate is accurate."""
    f_mc_mean = mean_fun(f_mc, 10_000)
    received, _num_nans = f_mc_mean(key)
    assert np.allclose(received, 1.0, rtol=1e-2)


@testing.parametrize("key", [prng.PRNGKey(1)])
@testing.parametrize("mean_fun1, mean_fun2", (_ALL_MEAN_FNS[1:], _ALL_MEAN_FNS[:-1]))
def test_mean_nested(f_mc, key, mean_fun1, mean_fun2):
    """Assert that nested mean-computation remains accurate."""
    f_mc_mean = mean_fun1(mean_fun2(f_mc, 5), 10_000)
    received, _num_nans = f_mc_mean(key)
    assert np.allclose(received, 1.0, rtol=1e-2)


@testing.parametrize("key", [prng.PRNGKey(1)])
@testing.parametrize("mean_fun", _ALL_MEAN_FNS)
def test_mean_many_nans(key, mean_fun):
    """Assert that NaNs are captured (in an all-NaN situtation)."""

    def f(x):
        return np.nan() * np.ones_like(x)

    f_mc = montecarlo.montecarlo(f, sample_fun=prng.normal)
    f_mc_mean = mean_fun(f_mc, 10_000)
    _received, num_nans = f_mc_mean(key)
    assert num_nans == 10_000


@testing.parametrize("key", [prng.PRNGKey(1)])
@testing.parametrize(
    "mean_fun1, mean_fun2, mean_fun3",
    zip(_ALL_MEAN_FNS[1:], _ALL_MEAN_FNS[:-1], _ALL_MEAN_FNS[1:]),
)
def test_mean_many_nans_nested(key, mean_fun1, mean_fun2, mean_fun3):
    """Assert that nested mean-computation captures NaNs (in an all-Nan situation)."""

    def f(x):
        return np.nan() * np.ones_like(x)

    n1, n2, n3 = 3, 4, 5
    f_mc = montecarlo.montecarlo(f, sample_fun=prng.normal)
    f_mc_mean = mean_fun3(mean_fun2(mean_fun1(f_mc, n1), n2), n3)
    _received, num_nans = f_mc_mean(key)
    assert num_nans == n1 * n2 * n3
